from pathlib import Path

import yaml

MANIFEST_FILENAMES = ("pixi.toml", "pyproject.toml")
CONFIG_FILENAMES = (*MANIFEST_FILENAMES, "environment.yml", "pixi.lock")


def find_project_dir(input_files: list[Path]) -> list[Path]:
    path_dir = set()
    for input_file in input_files:
        filename = input_file.name
        if filename not in CONFIG_FILENAMES:
            raise ValueError(f"Expected filename to be one of {CONFIG_FILENAMES}")
        path_dir.add(input_file.parent)
    return list(path_dir)


def get_manifest_path(path_dir: Path) -> Path:
    for manifest_filename in MANIFEST_FILENAMES:
        manifest_path = path_dir / manifest_filename
        if manifest_path.is_file():
            return manifest_path

    raise ValueError(f"Could not find manifest path on directory {path_dir}")


def load_environment_file(
    path_dir: Path,
    environment_file: str = "environment.yml",
    raise_exception: bool = True,
) -> dict | list | None:
    filepath = path_dir / environment_file
    try:
        with open(filepath) as file:
            return yaml.safe_load(file)
    except FileNotFoundError as err:
        if not raise_exception:
            return None
        raise err


def save_environment_file(
    data, path_dir: Path, environment_file: str = "environment.yml"
):
    filepath = path_dir / environment_file
    with open(filepath, mode="w") as file:
        yaml.dump(
            data,
            file,
            default_flow_style=False,
            allow_unicode=True,
            encoding="utf-8",
            indent=2,
            sort_keys=False,
        )



import argparse
import logging
import sys
from pathlib import Path

from pixi_sync_environment.io import (
    CONFIG_FILENAMES,
    find_project_dir,
    get_manifest_path,
    load_environment_file,
    save_environment_file,
)
from pixi_sync_environment.pixi_environment import (
    create_environment_dict_from_pixi,
)

# Configure logging
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(levelname)s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)
logger = logging.getLogger(__name__)


def get_parser() -> argparse.ArgumentParser:
    """Create an ArgumentParser for the compare_environments function."""
    parser = argparse.ArgumentParser(
        description="Compare and update conda environment files using pixi manifest",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "input_files",
        nargs="+",
        type=Path,
        help=f"Path to configuration files ({'/'.join(CONFIG_FILENAMES)})",
    )

    parser.add_argument(
        "--environment-file",
        type=str,
        default="environment.yml",
        help="Name of the environment file",
    )

    parser.add_argument(
        "--explicit",
        action="store_true",
        default=False,
        help="Use explicit package specifications",
    )

    parser.add_argument(
        "--name", type=str, default=None, help="Environment name (optional)"
    )

    parser.add_argument(
        "--prefix", type=str, default=None, help="Environment prefix path (optional)"
    )

    parser.add_argument(
        "--environment", type=str, default="default", help="Name of pixi environment"
    )

    parser.add_argument(
        "--include-pip-packages",
        action="store_true",
        default=False,
        help="Include pip packages in the environment",
    )

    parser.add_argument(
        "--no-include-conda-channels",
        action="store_false",
        dest="include_conda_channels",
        default=True,
        help="Exclude conda channels from the environment",
    )

    parser.add_argument(
        "--include-build",
        action="store_true",
        default=False,
        help="Include build information",
    )

    return parser


def pixi_sync_environment(
    path_dir: Path,
    environment: str = "default",
    environment_file: str = "environment.yml",
    explicit: bool = False,
    name: str | None = None,
    prefix: str | None = None,
    include_pip_packages: bool = False,
    include_conda_channels: bool = True,
    include_build: bool = False,
):
    current_environment_dict = load_environment_file(
        path_dir, environment_file, raise_exception=False
    )
    manifest_path = get_manifest_path(path_dir)
    new_environment_dict = create_environment_dict_from_pixi(
        manifest_path,
        environment,
        explicit=explicit,
        name=name,
        prefix=prefix,
        include_pip_packages=include_pip_packages,
        include_conda_channels=include_conda_channels,
        include_build=include_build,
    )
    if not current_environment_dict:
        logger.info(
            "Couldn't load environment file, writing to %s", path_dir / environment_file
        )
        save_environment_file(
            new_environment_dict, path_dir, environment_file=environment_file
        )
    elif current_environment_dict != new_environment_dict:
        logger.info("%s not in sync with environment", environment_file)
        save_environment_file(
            new_environment_dict, path_dir, environment_file=environment_file
        )
    else:
        logger.info("environment.yml file already in sync")


def main() -> None:
    args = get_parser().parse_args()
    project_dirs = find_project_dir(args.input_files)
    if not project_dirs:
        sys.exit(1)

    for dir in project_dirs:
        logger.info("Syncing environment for directory %s", dir)
        pixi_sync_environment(
            dir,
            environment=args.environment,
            environment_file=args.environment_file,
            explicit=args.explicit,
            name=args.name,
            prefix=args.prefix,
            include_pip_packages=args.include_pip_packages,
            include_conda_channels=args.no_include_conda_channels,  # Weird but correct
            include_build=args.include_build,
        )



from dataclasses import dataclass
from typing import Literal


@dataclass
class PackageInfo:
    name: str
    version: str
    size_bytes: int
    build: str | None
    kind: Literal["conda", "pypi"]
    source: str
    is_explicit: bool
    is_editable: bool | None = None

    @property
    def is_conda_package(self):
        return self.kind == "conda"

    @property
    def is_pypi_package(self):
        return self.kind == "pypi"

    def get_package_spec_str(self, include_build: bool = False) -> str:
        properties = [self.name, self.version]
        if include_build and self.build is not None:
            properties.append(self.build)
        return "=".join(properties)



import json
import logging
import subprocess
from pathlib import Path
from typing import Iterable

from pixi_sync_environment.package_info import PackageInfo

logger = logging.getLogger(__name__)


def get_pixi_packages(
    manifest_path: Path, environment: str | None = None, explicit: bool = False
) -> list[PackageInfo]:
    logger.info("Getting explicit packages from pixi")
    args = [
        "pixi",
        "list",
        "--manifest-path",
        str(manifest_path),
        "--json",
    ]
    if environment:
        args += ["--environment", environment]
    if explicit:
        args.append("--explicit")
    cmd = " ".join(args)
    logger.info("Running %s", cmd)
    try:
        result = subprocess.run(
            args,
            capture_output=True,
            text=True,
            check=True,
        )
    except subprocess.CalledProcessError as err:
        logger.error(err.output)
        logger.exception(err)
        raise err

    package_list = json.loads(result.stdout)
    return [PackageInfo(**package) for package in package_list]


def create_environment_dict_from_packages(
    packages: Iterable[PackageInfo],
    name: str | None = None,
    prefix: str | None = None,
    include_pip_packages: bool = False,
    include_conda_channels: bool = True,
    include_build: bool = False,
) -> dict:
    conda_packages = [package for package in packages if package.is_conda_package]
    pypi_packages = [package for package in packages if package.is_pypi_package]
    dependecies: list[str | dict] = [
        package.get_package_spec_str(include_build=include_build)
        for package in conda_packages
    ]
    if pypi_packages and include_pip_packages:
        pypi_package_specs = [
            package.get_package_spec_str(include_build=False)
            for package in pypi_packages
        ]
        dependecies.append({"pip": pypi_package_specs})
    environment_dict = {}
    if name is not None:
        environment_dict["name"] = name
    if prefix is not None:
        environment_dict["prefix"] = prefix
    if include_conda_channels:
        environment_dict["channels"] = list(
            {package.source for package in conda_packages}
        )
    environment_dict["dependencies"] = dependecies
    return environment_dict


def create_environment_dict_from_pixi(
    manifest_path: Path,
    environment: str,
    explicit: bool = False,
    name: str | None = None,
    prefix: str | None = None,
    include_pip_packages: bool = False,
    include_conda_channels: bool = True,
    include_build: bool = False,
) -> dict:
    packages = get_pixi_packages(manifest_path, environment, explicit=explicit)
    return create_environment_dict_from_packages(
        packages,
        name=name,
        prefix=prefix,
        include_pip_packages=include_pip_packages,
        include_conda_channels=include_conda_channels,
        include_build=include_build,
    )
